/*
 * Tencent is pleased to support the open source community by making BK-CI 蓝鲸持续集成平台 available.
 *
 * Copyright (C) 2019 THL A29 Limited, a Tencent company.  All rights reserved.
 *
 * BK-CI 蓝鲸持续集成平台 is licensed under the MIT license.
 *
 * A copy of the MIT License is included in this file.
 *
 *
 * Terms of the MIT License:
 * ---------------------------------------------------
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package com.tencent.devops.misc.dao

import com.tencent.devops.environment.pojo.enums.NodeStatus
import com.tencent.devops.environment.pojo.enums.NodeType
import com.tencent.devops.model.environment.tables.TNode
import com.tencent.devops.model.environment.tables.records.TNodeRecord
import org.jooq.DSLContext
import org.springframework.stereotype.Repository

@Repository
class EnvironmentNodeDao {
    fun get(dslContext: DSLContext, projectId: String, nodeId: Long): TNodeRecord? {
        with(TNode.T_NODE) {
            return dslContext.selectFrom(this)
                    .where(NODE_ID.eq(nodeId))
                    .and(PROJECT_ID.eq(projectId))
                    .fetchOne()
        }
    }

    fun updateNodeStatus(
        dslContext: DSLContext,
        id: Long,
        status: NodeStatus
    ) {
        with(TNode.T_NODE) {
            dslContext.update(this)
                    .set(NODE_STATUS, status.name)
                    .where(NODE_ID.eq(id))
                    .execute()
        }
    }

    fun batchDeleteNode(dslContext: DSLContext, projectId: String, nodeIds: List<Long>) {
        if (nodeIds.isEmpty()) {
            return
        }

        with(TNode.T_NODE) {
            dslContext.deleteFrom(this)
                    .where(PROJECT_ID.eq(projectId))
                    .and(NODE_ID.`in`(nodeIds))
                    .execute()
        }
    }

    fun listDevCloudNodesByTaskId(dslContext: DSLContext, taskId: Long): List<TNodeRecord> {
        with(TNode.T_NODE) {
            return dslContext.selectFrom(this)
                    .where(NODE_TYPE.eq(NodeType.DEVCLOUD.name)).and(TASK_ID.eq(taskId))
                    .fetch()
        }
    }

    fun deleteDevCloudNodesByTaskId(dslContext: DSLContext, taskId: Long) {
        with(TNode.T_NODE) {
            dslContext.deleteFrom(this)
                    .where(NODE_TYPE.eq(NodeType.DEVCLOUD.name)).and(TASK_ID.eq(taskId))
                    .and(NODE_STATUS.eq(NodeStatus.DELETED.name))
                    .execute()
        }
    }

    fun updateNode(dslContext: DSLContext, allCmdbNodes: TNodeRecord) {
        dslContext.batchUpdate(allCmdbNodes).execute()
    }

    fun listAllServerNodes(dslContext: DSLContext): List<TNodeRecord> {
        with(TNode.T_NODE) {
            return dslContext.selectFrom(this)
                .where(NODE_TYPE.`in`(NodeType.CC.name, NodeType.CMDB.name, NodeType.BCSVM.name, NodeType.OTHER.name, NodeType.DEVCLOUD.name))
                .fetch()
        }
    }

    fun listAllNodesByType(dslContext: DSLContext, nodeType: NodeType): List<TNodeRecord> {
        with(TNode.T_NODE) {
            return dslContext.selectFrom(this)
                .where(NODE_TYPE.eq(nodeType.name))
                .fetch()
        }
    }

    fun batchUpdateNode(dslContext: DSLContext, allCmdbNodes: List<TNodeRecord>) {
        if (allCmdbNodes.isEmpty()) {
            return
        }
        dslContext.batchUpdate(allCmdbNodes).execute()
    }
}

/**

ALTER TABLE T_NODE ADD COLUMN `IMAGE` varchar(512) NULL COMMENT '镜像';
ALTER TABLE T_NODE ADD COLUMN `TASK_ID` bigint(20) NULL COMMENT '任务ID';

**/