/*
 * Tencent is pleased to support the open source community by making BK-CI 蓝鲸持续集成平台 available.
 *
 * Copyright (C) 2019 THL A29 Limited, a Tencent company.  All rights reserved.
 *
 * BK-CI 蓝鲸持续集成平台 is licensed under the MIT license.
 *
 * A copy of the MIT License is included in this file.
 *
 *
 * Terms of the MIT License:
 * ---------------------------------------------------
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package com.tencent.devops.process.websocket.push

import com.fasterxml.jackson.databind.ObjectMapper
import com.tencent.devops.common.event.annotation.Event
import com.tencent.devops.common.event.dispatcher.pipeline.mq.MQ
import com.tencent.devops.common.redis.RedisOperation
import com.tencent.devops.common.service.utils.SpringContextUtil
import com.tencent.devops.common.websocket.dispatch.message.PipelineMessage
import com.tencent.devops.common.websocket.dispatch.message.SendMessage
import com.tencent.devops.common.websocket.dispatch.push.WebsocketPush
import com.tencent.devops.common.websocket.pojo.NotifyPost
import com.tencent.devops.common.websocket.pojo.WebSocketType
import com.tencent.devops.common.websocket.utils.PageUtils
import com.tencent.devops.common.websocket.utils.RedisUtlis
import com.tencent.devops.process.engine.service.PipelineService
import com.tencent.devops.process.pojo.PipelineStatus
import org.slf4j.LoggerFactory

@Event(exchange = MQ.EXCHANGE_WEBSOCKET_TMP_FANOUT, routeKey = MQ.ROUTE_WEBSOCKET_TMP_EVENT)
data class StatusWebsocketPush(
    val buildId: String?,
    val pipelineId: String,
    val projectId: String,
    override val userId: String,
    override val pushType: WebSocketType,
    override val redisOperation: RedisOperation,
    override val objectMapper: ObjectMapper,
    override var page: String?,
    override var notifyPost: NotifyPost
) : WebsocketPush(userId, pushType, redisOperation, objectMapper, page, notifyPost) {
    companion object {
        private val logger = LoggerFactory.getLogger(this::class.java)
        private val pipelineService = SpringContextUtil.getBean(PipelineService::class.java)
    }

    override fun findSession(page: String): List<String>? {
        if (page == "") {
            logger.warn("page empty: buildId[$buildId],projectId:[$projectId],pipelineId:[$pipelineId],page:[$page]")
        }

        val pageList = PageUtils.createAllTagPage(page!!)

        var sessionList = mutableListOf<String>()
        pageList.forEach {
            val redisSession = RedisUtlis.getSessionListFormPageSessionByPage(redisOperation, it)
            if (redisSession != null) {
                sessionList.addAll(redisSession)
            }
        }
        return sessionList
    }

    override fun buildMqMessage(): SendMessage? {
        return PipelineMessage(
                buildId = buildId,
                projectId = projectId,
                pipelineId = pipelineId,
                notifyPost = notifyPost,
                userId = userId,
                page = page,
                sessionList = findSession(page!!),
                startTime = System.currentTimeMillis()
        )
    }

    override fun buildNotifyMessage(message: SendMessage) {
        val notifyPost = message.notifyPost

        val currentTimestamp = System.currentTimeMillis()

        val status = pipelineService.getPipelineAllStatus(userId, projectId, pipelineId)

        if (status != null) {
            val result = status.map {
                it.pipelineId to PipelineStatus(
                        it.taskCount,
                        it.buildCount,
                        it.lock,
                        it.canManualStartup,
                        it.latestBuildStartTime,
                        it.latestBuildEndTime,
                        it.latestBuildStatus,
                        it.latestBuildNum,
                        it.latestBuildTaskName,
                        it.latestBuildEstimatedExecutionSeconds,
                        it.latestBuildId,
                        currentTimestamp,
                        it.runningBuildCount,
                        it.hasCollect
                )
            }.toMap()

            if (notifyPost != null) {
                notifyPost.message = objectMapper.writeValueAsString(result)
//                logger.info("StatusWebsocketPush message: $notifyPost")
            }
        }
    }
}